/*   FILE: pypvm_core_module.c -- A python interface to PVM.  For more
 *         information on PVM, see http://www.epm.ornl.gov/pvm/.
 * AUTHOR: W. Michael Petullo, wp0002@drake.edu
 *   DATE: 13 FEB 1998
 *
 * Copyright (c) 1999 W. Michael Petullo
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <python1.5/Python.h>
#include <pvm3.h>

#define RAISE_ERRNO { raise_errno (); return (NULL); }

static void *py_malloc(size_t size);
static PyObject *pvm_strerror();
static void raise_errno();
static PyObject *PypvmError;

/* ============================ Process Control ============================ */

/* ============================ Py_pvm_mytid () ============================ */
static PyObject *Py_pvm_mytid(PyObject * self, PyObject * args)
/* POST: FCTVAL == task identifier of this PVM task */
{
    int mytid;
    if (!PyArg_ParseTuple(args, ""))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((mytid = pvm_mytid()) < 0)
	RAISE_ERRNO;
    return (Py_BuildValue("i", mytid));
}

/* ============================ Py_pvm_exit () ============================= */
static PyObject *Py_pvm_exit(PyObject * self, PyObject * args)
{
    if (!PyArg_ParseTuple(args, ""))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if (pvm_exit() < 0)
	RAISE_ERRNO;
    Py_INCREF(Py_None);
    return (Py_None);
}

/* ============================ Py_pvm_spawn () ============================ */
static PyObject *Py_pvm_spawn(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       task  -- string containing the name of a task to spawn
 *       argv  -- a list of string arguments to task
 *       flag  -- an integer with encoded options (see PVM documentation)
 *       where -- a string containing where to spawn tasks (or "")
 *       ntask -- an integer > 0 representing the number of tasks to spawn
 * POST: FCTVAL == list containing TIDs of spawned tasks
 */
{
    char *task, *where;
    char **argv = (char **) 0;
    int ntask, numt, list_len, flag, *tids, i;
    PyObject *argument, *argv_list, *arg, *tid, *tids_list;
    if (!PyArg_ParseTuple
	(args, "sOisi", &task, &argv_list, &flag, &where, &ntask))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((list_len = PyList_Size(argv_list)) < 0) {
	/* ASSERT: second argument was not a list */
	PyErr_SetString(PyExc_TypeError,
			"argument 2: expected list of strings");
	return (NULL);
    }
    if (list_len) {
	/* allocate memory for argument array */
	if ((argv = (char **) py_malloc(list_len * sizeof(char *) + 1)) ==
	    NULL) return (NULL);
	argv[list_len] = NULL;
    }
    for (i = 0; i < list_len; i++) {
	/* build argument array */
	arg = PyList_GetItem(argv_list, i);
	if ((argv[i] = PyString_AsString(arg)) == NULL) {
	    PyErr_SetString(PyExc_TypeError,
			    "argument 2: expected list of strings");
	    return (NULL);
	}
    }
    if ((tids = (int *) py_malloc(ntask * sizeof(int))) == NULL)
	return (NULL);
    if ((numt = pvm_spawn(task, argv, flag, where, ntask, tids)) == 0)
	RAISE_ERRNO;
    tids_list = PyList_New(numt);
    for (i = 0; i < numt; i++)
	/* build list of tids to return to Python */
	PyList_SetItem(tids_list, i, PyInt_FromLong(tids[i]));
    free(tids);
/* free memory allocated for argument array */
    for (i = 0; i < list_len; i++)
	free(argv[i]);
    free(argv);
    return (tids_list);
}

/* ============================ Py_pvm_kill () ============================= */
static PyObject *Py_pvm_kill(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       tid -- an integer tid of a task to kill
 */
{
    int tid;
    if (!PyArg_ParseTuple(args, "i", &tid))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if (pvm_kill(tid) < 0)
	RAISE_ERRNO;
    Py_INCREF(Py_None);
    return (Py_None);
}

/* ============================ Py_pvm_catchout () ========================= */
static PyObject *Py_pvm_catchout(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       filemame -- a file to print catched task output to
 */
{
    FILE *file;
    PyObject *py_file;
    if (!PyArg_ParseTuple(args, "O", &py_file))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if (!PyFile_Check(py_file)) {
	PyErr_SetString(PyExc_TypeError, "argument 1: expected file");
	return (NULL);
    }
    file = PyFile_AsFile(py_file);
    if (pvm_catchout(file) < 0)
	RAISE_ERRNO;
    Py_INCREF(Py_None);
    return (Py_None);
}

/* ============================ Information ================================ */

/* ============================ Py_pvm_parent () =========================== */
static PyObject *Py_pvm_parent(PyObject * self, PyObject * args)
/* POST: FCTVAL == task identifier of parent PVM task */
{
    int parent;
    if (!PyArg_ParseTuple(args, ""))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((parent = (pvm_parent())) < 0)
	RAISE_ERRNO;
    return (PyInt_FromLong(parent));
}

/* ============================ Py_pvm_tidtohost () ======================== */
static PyObject *Py_pvm_tidtohost(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       tid -- an integer tid, call it x
 * POST: FCTVAL == tid of the deamon running on the same host as x
 */
{
    int tid, dtid;
    if (!PyArg_ParseTuple(args, "i", &tid))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((dtid = pvm_tidtohost(tid)) < 0)
	RAISE_ERRNO;
    return (Py_BuildValue("i", dtid));
}

/* ============================ Py_pvm_config () =========================== */
static PyObject *Py_pvm_config(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       nhosts >= the number of hosts (for memory allocation purposes)
 * POST: FCTVAL == number of hosts
 *              && number of data formats
 *              && tuple of information about the virtual machine
 * NOTE: the information is in the form of a Python dictionary containing:
 *              dtid:     w
 *              hostname: x
 *              arch:     y
 *              speed:    z
 */
{
    int nhosts, narchs, i;
    struct pvmhostinfo *info;
    PyObject *py_info, *items, *fctval;
    if (!PyArg_ParseTuple(args, "i", &nhosts))
	/* ASSERT: incorrect arguments */
	return (NULL);
    info =
	(struct pvmhostinfo *) malloc(nhosts * sizeof(struct pvmhostinfo));
    if (pvm_config(&nhosts, &narchs, &info) < 0)
	RAISE_ERRNO;
    py_info = PyList_New(nhosts);
    for (i = 0; i < nhosts; i++) {
	/* build info dictionaries */
	items = PyDict_New();
	PyDict_SetItemString(items, "dtid",
			     PyInt_FromLong(info[i].hi_tid));
	PyDict_SetItemString(items, "hostname",
			     PyString_FromString(info[i].hi_name));
	PyDict_SetItemString(items, "arch",
			     PyString_FromString(info[i].hi_arch));
	PyDict_SetItemString(items, "speed",
			     PyInt_FromLong(info[i].hi_speed));
	PyList_SetItem(py_info, i, items);
    }
    fctval = PyTuple_New(3);
    PyTuple_SetItem(fctval, 0, PyInt_FromLong(nhosts));
    PyTuple_SetItem(fctval, 1, PyInt_FromLong(narchs));
    PyTuple_SetItem(fctval, 2, py_info);
    free(info);
    return (fctval);
}

/* ============================ Py_pvm_tasks () ============================ */
static PyObject *Py_pvm_tasks(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       which  == 0, a pvmd tid, or a tid
 *       ntasks >= the number of tasks (for memory allocation purposes)
 * POST: FCTVAL == number of tasks
 *              && tuple of information about the virtual machine
 * NOTE: the information is in the form of a Python dictionary containing:
 *              ti_tid:   u
 *              ti_ptid:  v
 *              ti_host:  w
 *              ti_flag:  x
 *              ti_a_out: y
 *              ti_pid:   z
 */
{
    int which, ntasks, i;
    struct pvmtaskinfo *info;
    PyObject *py_info, *items, *fctval;
    if (!PyArg_ParseTuple(args, "ii", &which, &ntasks))
	/* ASSERT: incorrect arguments */
	return (NULL);
    info =
	(struct pvmtaskinfo *) malloc(ntasks * sizeof(struct pvmtaskinfo));
    if (pvm_tasks(which, &ntasks, &info) < 0)
	RAISE_ERRNO;
    py_info = PyList_New(ntasks);
    for (i = 0; i < ntasks; i++) {
	/* build info dictionaries */
	items = PyDict_New();
	PyDict_SetItemString(items, "ti_tid",
			     PyInt_FromLong(info[i].ti_tid));
	PyDict_SetItemString(items, "ti_ptid",
			     PyInt_FromLong(info[i].ti_ptid));
	PyDict_SetItemString(items, "ti_host",
			     PyInt_FromLong(info[i].ti_host));
	PyDict_SetItemString(items, "ti_flag",
			     PyInt_FromLong(info[i].ti_flag));
	PyDict_SetItemString(items, "ti_a_out",
			     PyString_FromString(info[i].ti_a_out));
	PyDict_SetItemString(items, "ti_pid",
			     PyInt_FromLong(info[i].ti_pid));
	PyList_SetItem(py_info, i, items);
    }
    fctval = PyTuple_New(2);
    PyTuple_SetItem(fctval, 0, PyInt_FromLong(ntasks));
    PyTuple_SetItem(fctval, 1, py_info);
    free(info);
    return (fctval);
}

/* ============================ Dynamic Configuration ====================== */

/* ============================ Py_pvm_addhosts () ========================= */
static PyObject *Py_pvm_addhosts(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       hosts -- a list of hosts to add to the virtual machine
 * POST: FCTVAL == list containing the status of each host
 */
{
    int num_hosts, i, *infos;
    char **hosts = (char **) 0;
    PyObject *host_list, *host, *info_list;
    if (!PyArg_ParseTuple(args, "O", &host_list))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((num_hosts = PyList_Size(host_list)) < 0) {
	/* ASSERT: second argument was not a list */
	PyErr_SetString(PyExc_TypeError,
			"argument 1: expected list of strings");
	return (NULL);
    }
    if (num_hosts)
	/* allocate memory for host array */
	if ((hosts = (char **) py_malloc(num_hosts * sizeof(char *) + 1))
	    == NULL)
	    return (NULL);
    for (i = 0; i < num_hosts; i++) {
	/* build host array */
	host = PyList_GetItem(host_list, i);
	if ((hosts[i] = PyString_AsString(host)) == NULL) {
	    PyErr_SetString(PyExc_TypeError,
			    "argument 1: expected list of strings");
	    return (NULL);
	}
    }
    if ((infos = (int *) py_malloc(num_hosts * sizeof(int))) == NULL)
	return (NULL);
    if (pvm_addhosts(hosts, num_hosts, infos) < 0)
	RAISE_ERRNO;
    info_list = PyList_New(num_hosts);
    for (i = 0; i < num_hosts; i++)
	PyList_SetItem(info_list, i, PyInt_FromLong(infos[i]));
    free(hosts);
    free(infos);
    return (info_list);
}

/* ============================ Py_pvm_delhosts () ========================= */
static PyObject *Py_pvm_delhosts(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       hosts -- a list of hosts to add to the virtual machine
 * POST: FCTVAL == list containing the status of each host deleted
 */
{
    int num_hosts, i, *infos;
    char **hosts = (char **) 0;
    PyObject *host_list, *host, *info_list;
    if (!PyArg_ParseTuple(args, "O", &host_list))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((num_hosts = PyList_Size(host_list)) < 0) {
	/* ASSERT: second argument was not a list */
	PyErr_SetString(PyExc_TypeError,
			"argument 1: expected list of strings");
	return (NULL);
    }
    if (num_hosts)
	/* allocate memory for host array */
	if ((hosts = (char **) py_malloc(num_hosts * sizeof(char *) + 1))
	    == NULL)
	    return (NULL);
    for (i = 0; i < num_hosts; i++) {
	/* build host array */
	host = PyList_GetItem(host_list, i);
	if ((hosts[i] = PyString_AsString(host)) == NULL) {
	    PyErr_SetString(PyExc_TypeError,
			    "argument 1: expected list of strings");
	    return (NULL);
	}
    }
    if ((infos = (int *) py_malloc(num_hosts * sizeof(int))) == NULL)
	return (NULL);
    if (pvm_delhosts(hosts, num_hosts, infos) < 0)
	RAISE_ERRNO;
    info_list = PyList_New(num_hosts);
    for (i = 0; i < num_hosts; i++)
	PyList_SetItem(info_list, i, PyInt_FromLong(infos[i]));
    free(hosts);
    free(infos);
    return (info_list);
}

/* ============================ Signaling ================================== */

/* ============================ Py_pvm_sendsig () ========================== */
static PyObject *Py_pvm_sendsig(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       tid    -- tid to send signal to
 *       signum -- number of signal to send
 */
{
    int tid, signum;
    if (!PyArg_ParseTuple(args, "ii", &tid, &signum))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if (pvm_sendsig(tid, signum) < 0)
	RAISE_ERRNO;
    Py_INCREF(Py_None);
    return (Py_None);
}

/* ============================ Py_pvm_notify () =========================== */
static PyObject *Py_pvm_notify(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       what   -- an integer type of event to trigger the notification
 *       msgtag -- an integer msgtag to be used in notification
 *       cnt    -- an integer count
 *       tids   -- a list of tids to be notified about
 */
{
    int what, msgtag, cnt, *tids, i;
    PyObject *tid_list, *item;
    if (!PyArg_ParseTuple(args, "iiiO", &what, &msgtag, &cnt, &tid_list))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if (what != PvmHostAdd) {
	/* tids not used for PvmHostAdd */
	if ((tids = (int *) py_malloc(cnt * sizeof(int))) == NULL)
	    return (NULL);
	for (i = 0; i < cnt; i++) {
	    item = PyList_GetItem(tid_list, i);
	    if (!PyInt_Check(item)) {
		/* ASSERT: second argument was not a list */
		PyErr_SetString(PyExc_TypeError,
				"argument 4: expected list of tids");
		return (NULL);
	    }
	    tids[i] = PyInt_AsLong(item);
	}
	if (pvm_notify(what, msgtag, cnt, tids) < 0)
	    RAISE_ERRNO;
	Py_INCREF(Py_None);
	return (Py_None);
    }
}


/* ============================ Setting and  Getting Options =============== */

/* ============================ Py_pvm_setopt () =========================== */
static PyObject *Py_pvm_setopt(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       what -- an integer specifying what option to set
 *       val  -- the value to set what to
 * POST: FCTVAL == previous value of what
 */
{
    int what, val, oldval;
    if (!PyArg_ParseTuple(args, "ii", &what, &val))
	/* incorrect arguments */
	return (NULL);
    if ((oldval = pvm_setopt(what, val)) == PvmBadParam)
	RAISE_ERRNO;
    return (PyLong_FromLong(oldval));
}
/* ============================ Py_pvm_getopt () =========================== */
static PyObject *Py_pvm_getopt(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       what -- an integer specifying what option to get the value of
 * POST: FCTVAL == pvm_getopt (val)
 */
{
    int what, val;
    if (!PyArg_ParseTuple(args, "i", &what))
	/* incorrect arguments */
	return (NULL);
    if ((val = pvm_getopt(what)) == PvmBadParam)
	RAISE_ERRNO;
    return (PyLong_FromLong(val));
}
/* ============================ Message Passing ============================ */
/* ============================ Py_pvm_initsend () ========================= */
static PyObject *Py_pvm_initsend(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       encoding -- an integer representing the encoding method to use for
 *                   communication
 * POST: FCTVAL == bufid of created message buffer
 * NOTE: The encoding method mappings are as follows:
 *       PvmDataDefault = 0
 *       PvmDataRaw     = 1
 *       PvmDataInPlace = 2
 *       See the PVM documentation for more information.
 */
{
    int encoding;
    if (!PyArg_ParseTuple(args, "i", &encoding))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if (pvm_initsend(encoding) < 0)
	RAISE_ERRNO;
    Py_INCREF(Py_None);
    return (Py_None);
}
/* ============================ Py_pvm_mkbuf () ============================ */
static PyObject *Py_pvm_mkbuf(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       encoding -- an integer representing the encoding method to use for
 *                   communication
 * POST: FCTVAL == bufid of created message buffer
 * NOTE: The encoding method mappings are as follows:
 *       PvmDataDefault = 0
 *       PvmDataRaw     = 1
 *       PvmDataInPlace = 2
 *       See the PVM documentation for more information.
 */
{
    int encoding, newbuf;
    if (!PyArg_ParseTuple(args, "i", &encoding))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((newbuf = pvm_mkbuf(encoding)) < 0)
	RAISE_ERRNO;
    return (PyInt_FromLong(newbuf));
}
/* ============================ Py_pvm_freebuf () ========================== */
static PyObject *Py_pvm_freebuf(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       bufid -- an integer
 */
{
    int bufid;
    if (!PyArg_ParseTuple(args, "i", &bufid))
	/* incorrect arguments */
	return (NULL);
    if (pvm_freebuf(bufid) < 0)
	RAISE_ERRNO;
    Py_INCREF(Py_None);
    return (Py_None);
}
/* ============================ Py_pvm_getsbuf () ========================== */
static PyObject *Py_pvm_getsbuf(PyObject * self, PyObject * args)
/* POST: FCTVAL == bufid of active send buffer */
{
    int bufid;
    if (!PyArg_ParseTuple(args, ""))
	/* incorrect arguments */
	return (NULL);
    if ((bufid = pvm_getsbuf()) < 0)
	RAISE_ERRNO;
    return (PyInt_FromLong(bufid));
}
/* ============================ Py_pvm_getrbuf () ========================== */
static PyObject *Py_pvm_getrbuf(PyObject * self, PyObject * args)
/* POST: FCTVAL == bufid of active receive buffer */
{
    int bufid;
    if (!PyArg_ParseTuple(args, ""))
	/* incorrect arguments */
	return (NULL);
    if ((bufid = pvm_getrbuf()) < 0)
	RAISE_ERRNO;
    return (PyInt_FromLong(bufid));
}
/* ============================ Py_pvm_setsbuf () ========================== */
static PyObject *Py_pvm_setsbuf(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       bufid -- an integer
 */
{
    int bufid;
    if (!PyArg_ParseTuple(args, "i", &bufid))
	/* incorrect arguments */
	return (NULL);
    if (pvm_setsbuf(bufid) < 0)
	RAISE_ERRNO;
    Py_INCREF(Py_None);
    return (Py_None);
}
/* ============================ Py_pvm_setrbuf () ========================== */
static PyObject *Py_pvm_setrbuf(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       bufid -- an integer
 */
{
    int bufid;
    if (!PyArg_ParseTuple(args, "i", &bufid))
	/* incorrect arguments */
	return (NULL);
    if (pvm_setrbuf(bufid) < 0)
	RAISE_ERRNO;
    Py_INCREF(Py_None);
    return (Py_None);
}
/* ============================ Py_pvm_pkbyte () =========================== */
static PyObject *Py_pvm_pkbyte(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       char_list -- a list of chars to pack
 *       stride   -- an integer stride to use when packing
 */
{
    char *chars;
    int stride, list_len, i;
    PyObject *char_list, *item;
    if (!PyArg_ParseTuple(args, "Oi", &char_list, &stride))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((list_len = PyList_Size(char_list)) < 0) {
	/* ASSERT: second argument was not a list */
	PyErr_SetString(PyExc_TypeError,
			"argument 2: expected list of chars");
	return (NULL);
    }
    if ((chars = (char *) py_malloc(list_len * sizeof(char))) < 0)
	return (NULL);
    for (i = 0; i < list_len; i++) {
	/* extract chars from list */
	item = PyList_GetItem(char_list, i);
	if ((chars[i] = PyString_AsString(item)[0]) < 0
	    || PyString_Size(item) > 1) {
	    /* not a list of strings or sting is more than one character */
	    PyErr_SetString(PyExc_TypeError,
			    "argument 2: expected list of chars");
	    return (NULL);
	}
    }
    if (pvm_pkbyte(chars, list_len, stride) < 0)
	RAISE_ERRNO;
    free(chars);
    Py_INCREF(Py_None);
    return (Py_None);
}

/* ============================ Py_pvm_pkdouble () ========================= */
static PyObject *Py_pvm_pkdouble(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       double_list -- a list of doubles to pack
 *       stride   -- an integer stride to use when packing
 */
{
    int stride, list_len, i;
    double *doubles;
    PyObject *double_list, *item;
    if (!PyArg_ParseTuple(args, "Oi", &double_list, &stride))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((list_len = PyList_Size(double_list)) < 0) {
	/* ASSERT: second argument was not a list */
	PyErr_SetString(PyExc_TypeError,
			"argument 2: expected list of doubles");
	return (NULL);
    }
    if ((doubles = (double *) py_malloc(list_len * sizeof(double))) ==
	NULL) return (NULL);
    for (i = 0; i < list_len; i++) {
	/* extract doubles from list */
	item = PyList_GetItem(double_list, i);
	if (!PyFloat_Check(item)) {
	    PyErr_SetString(PyExc_TypeError,
			    "argument 2: expected list of doubles");
	    return NULL;
	}
	doubles[i] = PyFloat_AsDouble(item);
    }
    if (pvm_pkdouble(doubles, list_len, stride) < 0)
	RAISE_ERRNO;
    free(doubles);
    Py_INCREF(Py_None);
    return (Py_None);
}

/* ============================ Py_pvm_pkfloat () ========================== */
static PyObject *Py_pvm_pkfloat(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       float_list -- a list of floats to pack
 *       stride   -- an integer stride to use when packing
 */
{
    int stride, list_len, i;
    float *floats;
    PyObject *float_list, *item;
    if (!PyArg_ParseTuple(args, "Oi", &float_list, &stride))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((list_len = PyList_Size(float_list)) < 0) {
	/* ASSERT: second argument was not a list */
	PyErr_SetString(PyExc_TypeError,
			"argument 2: expected list of floats");
	return (NULL);
    }
    if ((floats = (float *) py_malloc(list_len * sizeof(float))) == NULL)
	return (NULL);
    for (i = 0; i < list_len; i++) {
	/* extract floats from list */
	item = PyList_GetItem(float_list, i);
	if (!PyInt_Check(item)) {
	    PyErr_SetString(PyExc_TypeError,
			    "argument 2: expected list of floats");
	    return NULL;
	}
	floats[i] = PyInt_AsLong(item);
    }
    if (pvm_pkfloat(floats, list_len, stride) < 0)
	RAISE_ERRNO;
    free(floats);
    Py_INCREF(Py_None);
    return (Py_None);
}

/* ============================ Py_pvm_pkint () ============================ */
static PyObject *Py_pvm_pkint(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       int_list -- a list of integers to pack
 *       stride   -- an integer stride to use when packing
 */
{
    int stride, list_len, *ints, i;
    PyObject *int_list, *item;
    if (!PyArg_ParseTuple(args, "Oi", &int_list, &stride))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((list_len = PyList_Size(int_list)) < 0) {
	/* ASSERT: second argument was not a list */
	PyErr_SetString(PyExc_TypeError,
			"argument 2: expected list of ints");
	return (NULL);
    }
    if ((ints = (int *) py_malloc(list_len * sizeof(int))) == NULL)
	return (NULL);
    for (i = 0; i < list_len; i++) {
	/* extract integers from list */
	item = PyList_GetItem(int_list, i);
	if (!PyInt_Check(item)) {
	    PyErr_SetString(PyExc_TypeError,
			    "argument 2: expected list of ints");
	    return NULL;
	}
	ints[i] = (int) PyInt_AsLong(item);
    }
    if (pvm_pkint(ints, list_len, stride) < 0)
	RAISE_ERRNO;
    free(ints);
    Py_INCREF(Py_None);
    return (Py_None);
}

/* ============================ Py_pvm_pklong () =========================== */
static PyObject *Py_pvm_pklong(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       long_list -- a list of longs to pack
 *       stride   -- an integer stride to use when packing
 */
{
    int stride, list_len, i;
    long *longs;
    PyObject *long_list, *item;
    if (!PyArg_ParseTuple(args, "Oi", &long_list, &stride))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((list_len = PyList_Size(long_list)) < 0) {
	/* ASSERT: second argument was not a list */
	PyErr_SetString(PyExc_TypeError,
			"argument 2: expected list of longs");
	return (NULL);
    }
    if ((longs = (long *) py_malloc(list_len * sizeof(long))) == NULL)
	return (NULL);
    for (i = 0; i < list_len; i++) {
	/* extract longs from list */
	item = PyList_GetItem(long_list, i);
	if (!PyInt_Check(item)) {
	    PyErr_SetString(PyExc_TypeError,
			    "argument 2: expected list of longs");
	    return NULL;
	}
	longs[i] = PyInt_AsLong(item);
    }
    if (pvm_pklong(longs, list_len, stride) < 0)
	RAISE_ERRNO;
    free(longs);
    Py_INCREF(Py_None);
    return (Py_None);
}

/* ============================ Py_pvm_pkshort () ========================== */
static PyObject *Py_pvm_pkshort(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       short_list -- a list of shorts to pack
 *       stride   -- an integer stride to use when packing
 */
{
    short *shorts;
    int stride, list_len, i;
    PyObject *short_list, *item;
    if (!PyArg_ParseTuple(args, "Oi", &short_list, &stride))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((list_len = PyList_Size(short_list)) < 0) {
	/* ASSERT: second argument was not a list */
	PyErr_SetString(PyExc_TypeError,
			"argument 2: expected list of shortss");
	return (NULL);
    }
    if ((shorts = (short *) py_malloc(list_len * sizeof(short))) == NULL)
	return (NULL);
    for (i = 0; i < list_len; i++) {
	/* extract shorts from list */
	item = PyList_GetItem(short_list, i);
	if (!PyInt_Check(item)) {
	    PyErr_SetString(PyExc_TypeError,
			    "argument 2: expected list of shorts");
	    return NULL;
	}
	shorts[i] = PyInt_AsLong(item);
    }
    if (pvm_pkshort(shorts, list_len, stride) < 0)
	RAISE_ERRNO;
    free(shorts);
    Py_INCREF(Py_None);
    return (Py_None);
}

/* ============================ Py_pvm_pkstr () ============================ */
static PyObject *Py_pvm_pkstr(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       str    -- a string to pack
 */
{
    char *str;
    if (!PyArg_ParseTuple(args, "s", &str))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if (pvm_pkstr(str) < 0)
	RAISE_ERRNO;
    Py_INCREF(Py_None);
    return (Py_None);
}
/* ============================ Py_pvm_send () ============================= */
static PyObject *Py_pvm_send(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       tid    -- an integer tid to send the message to
 *       msgtag -- an integer identifier for the message
 */
{
    int tid, msgtag;
    if (!PyArg_ParseTuple(args, "ii", &tid, &msgtag))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if (pvm_send(tid, msgtag) < 0)
	RAISE_ERRNO;
    Py_INCREF(Py_None);
    return (Py_None);
}
/* ============================ Py_pvm_mcast () ============================ */
static PyObject *Py_pvm_mcast(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       tids   -- a list of tids to send the message to
 *       msgtag -- an integer identifier for the message
 */
{
    PyObject *tids_list, *tid;
    int *tids, msgtag, list_len, i;
    if (!PyArg_ParseTuple(args, "Oi", &tids_list, &msgtag))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((list_len = PyList_Size(tids_list)) < 0) {
	/* ASSERT: second argument was not a list */
	PyErr_SetString(PyExc_TypeError,
			"argument 1: expected list of tids");
	return (NULL);
    }
    if ((tids = (int *) py_malloc(list_len * sizeof(int))) == NULL)
	return (NULL);
    for (i = 0; i < list_len; i++) {
	tid = PyList_GetItem(tids_list, i);
	/* extract tids from list */
	if (!PyInt_Check(tid)) {
	    PyErr_SetString(PyExc_TypeError,
			    "argument 1: expected list of tids");
	    return (NULL);
	}
	tids[i] = PyInt_AsLong(tid);

    }
    if (pvm_mcast(tids, list_len, msgtag) < 0)
	RAISE_ERRNO;
    free(tids);
    Py_INCREF(Py_None);
    return (Py_None);
}

/* ============================ Py_pvm_recv () ============================= */
static PyObject *Py_pvm_recv(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       tid    -- an integer tid to receive the message from
 *       msgtag -- an integer identifier for the message
 * POST: FCTVAL == bufid of message buffer
 */
{
    int tid, msgtag, bufid;
    if (!PyArg_ParseTuple(args, "ii", &tid, &msgtag))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((bufid = pvm_recv(tid, msgtag)) < 0)
	RAISE_ERRNO;
    return (PyInt_FromLong(bufid));
}
/* ============================ Py_pvm_nrecv () ============================ */
static PyObject *Py_pvm_nrecv(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       tid    -- an integer TID
 *       msgtag -- an integer message tag
 * POST: FCTVAL == bufid of message buffer
 */
{
    int tid, msgtag, bufid;
    if (!PyArg_ParseTuple(args, "ii", &tid, &msgtag))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((bufid = pvm_nrecv(tid, msgtag)) < 0)
	RAISE_ERRNO;
    return (PyInt_FromLong(bufid));
}
/* ============================ Py_pvm_probe () ============================ */
static PyObject *Py_pvm_probe(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       tid    -- an integer tid to receive the message from
 *       msgtag -- an integer identifier for the message
 * POST: FCTVAL == bufid of message buffer
 */
{
    int tid, msgtag, bufid;
    if (!PyArg_ParseTuple(args, "ii", &tid, &msgtag))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((bufid = pvm_probe(tid, msgtag)) < 0)
	RAISE_ERRNO;
    return (PyInt_FromLong(bufid));
}
/* ============================ Py_pvm_trecv () ============================ */
static PyObject *Py_pvm_trecv(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       tid    -- an integer tid to receive the message from
 *       msgtag -- an integer identifier for the message
 *       secs   -- a time in seconds to give up after
 */
{
    int tid, msgtag, secs, bufid;
    struct timeval timeout;
    if (!PyArg_ParseTuple(args, "iii", &tid, &msgtag, &secs))
	/* ASSERT: incorrect arguments */
	return (NULL);
    timeout.tv_sec = secs;
    timeout.tv_usec = secs;
    if ((bufid = pvm_trecv(tid, msgtag, &timeout)) < 0)
	RAISE_ERRNO;
    return (PyInt_FromLong(bufid));
}
/* ============================ Py_pvm_bufinfo () ========================== */
static PyObject *Py_pvm_bufinfo(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       bufid -- an integer specifying a message buffer identifier
 * POST: FCTVAL == length in bytes of a message
 *              && message label
 *              && source of message
 */
{
    PyObject *fctval;
    int buff_id, bytes, msgtag, tid;
    if (!PyArg_ParseTuple(args, "i", &buff_id))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if (pvm_bufinfo(buff_id, &bytes, &msgtag, &tid) < 0)
	RAISE_ERRNO;
    fctval = PyTuple_New(3);
    PyTuple_SetItem(fctval, 0, PyInt_FromLong(bytes));
    PyTuple_SetItem(fctval, 1, PyInt_FromLong(msgtag));
    PyTuple_SetItem(fctval, 2, PyInt_FromLong(tid));
    return (fctval);
}
/* ============================ Py_pvm_upkbyte () ========================== */
static PyObject *Py_pvm_upkbyte(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       nitem  -- the number of items to unpack
 *       stride -- an integer stride to use when unpacking
 * POST: FCTVAL == byte encoded info from pvm_pkbyte call
 *              && a list containing the unpacked data
 */
{
    char *bytes;
    int nitem, stride, i;
    PyObject *byte_list, *tuple;
    if (!PyArg_ParseTuple(args, "ii", &nitem, &stride))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((bytes = (char *) py_malloc(nitem * sizeof(char))) == NULL)
	return (NULL);
    if (pvm_upkbyte(bytes, nitem, stride) < 0)
	RAISE_ERRNO;
    byte_list = PyList_New(nitem);
    for (i = 0; i < nitem; i++)
	/* build list of bytes to return to Python */
	PyList_SetItem(byte_list, i,
		       PyString_FromStringAndSize(&bytes[i], 1));
    free(bytes);
    return (byte_list);
}
/* ============================ Py_pvm_upkdouble () ======================== */
static PyObject *Py_pvm_upkdouble(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       nitem  -- the number of items to unpack
 *       stride -- an integer stride to use when unpacking
 * POST: FCTVAL == double encoded info from pvm_pkdouble call
 *              && a list containing the unpacked data
 */
{
    int nitem, stride, i;
    double *doubles;
    PyObject *double_list, *tuple;
    if (!PyArg_ParseTuple(args, "ii", &nitem, &stride))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((doubles = (double *) py_malloc(nitem * sizeof(double))) == NULL)
	return (NULL);
    if (pvm_upkdouble(doubles, nitem, stride) < 0)
	RAISE_ERRNO;
    double_list = PyList_New(nitem);
    for (i = 0; i < nitem; i++)
	/* build list of doubles to return to Python */

	PyList_SetItem(double_list, i, PyFloat_FromDouble(doubles[i]));
    free(doubles);
    return (double_list);
}
/* ============================ Py_pvm_upkfloat () ========================= */
static PyObject *Py_pvm_upkfloat(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       nitem  -- the number of items to unpack
 *       stride -- an integer stride to use when unpacking
 * POST: FCTVAL == float encoded info from pvm_pkfloat call
 *              && a list containing the unpacked data
 */
{
    int nitem, stride, i;
    float *floats;
    PyObject *float_list, *tuple;
    if (!PyArg_ParseTuple(args, "ii", &nitem, &stride))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((floats = (float *) py_malloc(nitem * sizeof(float)))
	== NULL)
	return (NULL);
    if (pvm_upkfloat(floats, nitem, stride) < 0)
	RAISE_ERRNO;
    float_list = PyList_New(nitem);
    for (i = 0; i < nitem; i++)
	/* build list of floats to return to Python */

	PyList_SetItem(float_list, i, PyFloat_FromDouble(floats[i]));
    free(floats);
    return (float_list);
}
/* ============================ Py_pvm_upkint () =========================== */
static PyObject *Py_pvm_upkint(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       nitem  -- the number of items to unpack
 *       stride -- an integer stride to use when unpacking
 * POST: FCTVAL == integer encoded info from pvm_pkint call
 *              && a list containing the unpacked data
 */
{
    int nitem, stride, *ints, i;
    PyObject *int_list, *tuple;
    if (!PyArg_ParseTuple(args, "ii", &nitem, &stride))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((ints = (int *) py_malloc(nitem * sizeof(int))) == NULL)
	return (NULL);
    if (pvm_upkint(ints, nitem, stride) < 0)
	RAISE_ERRNO;
    int_list = PyList_New(nitem);
    for (i = 0; i < nitem; i++)
	/* build list of ints to return to Python */
	PyList_SetItem(int_list, i, Py_BuildValue("i", (ints[i])));
    free(ints);
    return (int_list);
}
/* ============================ Py_pvm_upklong () ========================== */
static PyObject *Py_pvm_upklong(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       nitem  -- the number of items to unpack
 *       stride -- an integer stride to use when unpacking
 * POST: FCTVAL == long encoded info from pvm_pklong call
 *              && a list containing the unpacked data
 */
{
    int nitem, stride, i;
    long *longs;
    PyObject *long_list, *tuple;
    if (!PyArg_ParseTuple(args, "ii", &nitem, &stride))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((longs = (long *) py_malloc(nitem * sizeof(long))) == NULL)
	return (NULL);
    if (pvm_upklong(longs, nitem, stride) < 0)
	RAISE_ERRNO;
    long_list = PyList_New(nitem);
    for (i = 0; i < nitem; i++)
	/* build list of longs to return to Python */
	PyList_SetItem(long_list, i, PyLong_FromLong(longs[i]));
    free(longs);
    return (long_list);
}
/* ============================ Py_pvm_upkshort () ========================= */
static PyObject *Py_pvm_upkshort(PyObject * self, PyObject * args)
/*  PRE: args contains the following Python objects:
 *       nitem  -- the number of items to unpack
 *       stride -- an integer stride to use when unpacking
 * POST: FCTVAL == short encoded info from pvm_pkint call
 *              && a list containing the unpacked data
 */
{
    short *shorts;
    int nitem, stride, i;
    PyObject *short_list, *tuple;
    if (!PyArg_ParseTuple(args, "ii", &nitem, &stride))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if ((shorts = (short *) py_malloc(nitem * sizeof(short)))
	== NULL)
	return (NULL);
    if (pvm_upkshort(shorts, nitem, stride) < 0)
	RAISE_ERRNO;
    short_list = PyList_New(nitem);
    for (i = 0; i < nitem; i++)
	/* build list of shorts to return to Python */
	PyList_SetItem(short_list, i, PyInt_FromLong(shorts[i]));
    free(shorts);
    return (short_list);
}
/* ============================ Py_pvm_upkstr () =========================== */
static PyObject *Py_pvm_upkstr(PyObject * self, PyObject * args)
/* POST: FCTVAL == a string containing the unpacked data */
{
    int bytes, msgtag, tid;
    char *str;
    if (!PyArg_ParseTuple(args, ""))
	/* ASSERT: incorrect arguments */
	return (NULL);
    if (pvm_bufinfo(pvm_getrbuf(), &bytes, &msgtag, &tid) < 0)
	RAISE_ERRNO;
/* currently mallocs size of entire buffer, not necissarily just string */
    str = (char *) malloc(bytes * sizeof(char) + 1);
    if (pvm_upkstr(str) < 0)
	RAISE_ERRNO;
    return (PyString_FromString(str));
}
/* ============================ Py_pvm_tuperror () ========================= */
static PyObject *Py_pvm_tuperror(PyObject * self, PyObject * args)
/* POST: FCTVAL == a tuple containing errno and the message associated 
 *                 with pvm_errno 
 */
{
    return (pvm_strerror());
}
/* ============================ py_malloc () =============================== */
static void *py_malloc(size_t size)
/*  PRE: size is assigned the ammount of memory to allocate
 * POST: FCTVAL == pointer to allocated memory
 *              || NULL if allocation failed
 * NOTE: An exception is raised if operation fails.  Calling function must be
 *       sure to return NULL to Python so that exception is handled.
 */
{
    void *ptr;
    if ((ptr = malloc(size)) == NULL) {
	PyErr_NoMemory();
	return (NULL);
    }
    return (ptr);
}

/* ============================ pvm_strerror () ============================ */
static PyObject *pvm_strerror()
/* POST: message associated with current value of pvm_errno is returned */
{
    extern int pvm_errno;
    extern char *pvm_errlist[];
    PyObject *error;
    error = PyTuple_New(2);
    PyTuple_SetItem(error, 0, PyInt_FromLong(pvm_errno));
    PyTuple_SetItem(error, 1,
		    PyString_FromString(pvm_errlist[-pvm_errno]));
    return (error);
}
/* ============================ raise_errno () ============================= */
static void raise_errno()
/* POST: current value of pvm_errno is raised as a Python exception
 * NOTE: Calling function must be sure to return NULL to Python.
 */
{
    PyErr_SetObject(PypvmError, pvm_strerror());
}

/* ============================ Method Registration Table ================== */
static struct PyMethodDef pypvm_core_methods[] = {
    {"mytid", Py_pvm_mytid, METH_VARARGS},
    {"exit", Py_pvm_exit, METH_VARARGS},
    {"spawn", Py_pvm_spawn, METH_VARARGS},
    {"kill", Py_pvm_kill, METH_VARARGS},
    {"catchout", Py_pvm_catchout, METH_VARARGS},
    {"initsend", Py_pvm_initsend, METH_VARARGS},
    {"send", Py_pvm_send, METH_VARARGS},
    {"recv", Py_pvm_recv, METH_VARARGS},
    {"parent", Py_pvm_parent, METH_VARARGS},
    {"mcast", Py_pvm_mcast, METH_VARARGS},
    {"pkstr", Py_pvm_pkstr, METH_VARARGS},
    {"upkstr", Py_pvm_upkstr, METH_VARARGS},
    {"bufinfo", Py_pvm_bufinfo, METH_VARARGS},
    {"probe", Py_pvm_probe, METH_VARARGS},
    {"tuperror", Py_pvm_tuperror, METH_VARARGS},
    {"tidtohost", Py_pvm_tidtohost, METH_VARARGS},
    {"config", Py_pvm_config, METH_VARARGS},
    {"tasks", Py_pvm_tasks, METH_VARARGS},
    {"sendsig", Py_pvm_sendsig, METH_VARARGS},
    {"addhosts", Py_pvm_addhosts, METH_VARARGS},
    {"delhosts", Py_pvm_addhosts, METH_VARARGS},
    {"notify", Py_pvm_notify, METH_VARARGS},
    {"setopt", Py_pvm_setopt, METH_VARARGS},
    {"getopt", Py_pvm_getopt, METH_VARARGS},
    {"mkbuf", Py_pvm_mkbuf, METH_VARARGS},
    {"getsbuf", Py_pvm_getsbuf, METH_VARARGS},
    {"getrbuf", Py_pvm_getrbuf, METH_VARARGS},
    {"setsbuf", Py_pvm_setsbuf, METH_VARARGS},
    {"setrbuf", Py_pvm_setrbuf, METH_VARARGS},
    {"freebuf", Py_pvm_freebuf, METH_VARARGS},
    {"pkbyte", Py_pvm_pkbyte, METH_VARARGS},
    {"pkcplx", Py_pvm_pkfloat, METH_VARARGS},
    {"pkdcplx", Py_pvm_pkdouble, METH_VARARGS},
    {"pkdouble", Py_pvm_pkdouble, METH_VARARGS},
    {"pkfloat", Py_pvm_pkfloat, METH_VARARGS},
    {"pkint", Py_pvm_pkint, METH_VARARGS},
    {"pklong", Py_pvm_pklong, METH_VARARGS},
    {"pkshort", Py_pvm_pkshort, METH_VARARGS},
    {"upkbyte", Py_pvm_upkbyte, METH_VARARGS},
    {"upkcplx", Py_pvm_upkfloat, METH_VARARGS},
    {"upkdcplx", Py_pvm_upkdouble, METH_VARARGS},
    {"upkdouble", Py_pvm_upkdouble, METH_VARARGS},
    {"upkfloat", Py_pvm_upkfloat, METH_VARARGS},
    {"upkint", Py_pvm_upkint, METH_VARARGS},
    {"upklong", Py_pvm_upklong, METH_VARARGS},
    {"upkshort", Py_pvm_upkshort, METH_VARARGS},
    {"nrecv", Py_pvm_nrecv, METH_VARARGS},
    {"trecv", Py_pvm_trecv, METH_VARARGS},
    {"nrecv", Py_pvm_nrecv, METH_VARARGS},
    {"trecv", Py_pvm_trecv, METH_VARARGS},
    {NULL, NULL}
};

/* ============================ Initialization Function ==================== */
void initpypvm_core()
{
    PyObject *m, *d;
    m = Py_InitModule("pypvm_core", pypvm_core_methods);
    d = PyModule_GetDict(m);
    PypvmError = Py_BuildValue("s", "pypvm.error");
    PyDict_SetItemString(d, "error", PypvmError);
    if (PyErr_Occurred())
	Py_FatalError("can't initialize module pypvm");
}
